@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2002-2022 ProcessOne, SARL. All Rights Reserved.
+ * Copyright (C) 2002-2024 ProcessOne, SARL. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -20,9 +20,11 @@
 #include <erl_nif.h>
 #include <openssl/err.h>
 #include <openssl/ssl.h>
+#include <openssl/pkcs12.h>
 #if OPENSSL_VERSION_NUMBER >= 0x30000000L
 #include <openssl/core_names.h>
 #include <openssl/decoder.h>
+#include <openssl/provider.h>
 #endif
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -72,6 +74,10 @@ typedef unsigned __int32 uint32_t;
 #define DH_set0_pqg(dh, dh_p, param, dh_g) (dh)->p = dh_p; (dh)->g = dh_g
 #endif
 
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
+#define SSL_get_extms_support(s) 0
+#endif
+
 #if OPENSSL_VERSION_NUMBER < 0x10002000L
 #define SSL_is_server(s) (s)->server
 #endif
@@ -415,6 +421,7 @@ static int setup_dh(SSL_CTX *ctx, const unsigned char *dh_der, size_t dh_size, c
         if (EVP_PKEY_fromdata_init(pctx) <= 0 ||
             EVP_PKEY_fromdata(pctx, &pkey, EVP_PKEY_KEY_PARAMETERS, params) <= 0) {
             EVP_PKEY_CTX_free(pctx);
+            return 0;
         }
         if (SSL_CTX_set0_tmp_dh_pkey(ctx, pkey) != 1) {
             EVP_PKEY_free(pkey);
@@ -969,7 +976,7 @@ get_decrypted_data(ErlNifEnv *env, state_t* state, int bytes_to_read, ERL_NIF_TE
 }
 
 static ERL_NIF_TERM
-return_read_write(ErlNifEnv *env, state_t* state, int bytes_to_read) {
+return_read_write(ErlNifEnv *env, state_t* state, int bytes_to_read, ERL_NIF_TERM tag) {
     ERL_NIF_TERM read;
     if (get_decrypted_data(env, state, bytes_to_read, &read) == 2) {
         enif_mutex_unlock(state->mtx);
@@ -979,7 +986,7 @@ return_read_write(ErlNifEnv *env, state_t* state, int bytes_to_read) {
 
     enif_mutex_unlock(state->mtx);
 
-    return enif_make_tuple2(env, write, read);
+    return enif_make_tuple3(env, tag, write, read);
 }
 
 static int
@@ -1095,23 +1102,27 @@ loop_nif(ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[]) {
                 if (res == 2) {
                     return err_term;
                 }
-                return return_read_write(env, state, bytes_to_read);
+                return return_read_write(env, state, bytes_to_read, enif_make_atom(env, "init"));
             } else {
-                enif_mutex_unlock(state->mtx);
+                res = do_send_queue(env, state, &err_term, &to_send);
+                if (res == 2) {
+                    return err_term;
+                }
                 int reason = ERR_GET_REASON(ERR_peek_error());
                 if (reason == SSL_R_DATA_LENGTH_TOO_LONG ||
                     reason == SSL_R_PACKET_LENGTH_TOO_LONG ||
                     reason == SSL_R_UNKNOWN_PROTOCOL ||
                     reason == SSL_R_UNEXPECTED_MESSAGE ||
                     reason == SSL_R_WRONG_VERSION_NUMBER ||
-		    reason == SSL_R_HTTP_REQUEST ||
-		    reason == SSL_R_HTTPS_PROXY_REQUEST)
+                    reason == SSL_R_HTTP_REQUEST ||
+                    reason == SSL_R_HTTPS_PROXY_REQUEST)
                     /* Do not report badly formed Client Hello */
-                    return ERR_T(enif_make_atom(env, "closed"));
+                    err_term = ERR_T(enif_make_atom(env, "closed"));
                 else if (state->sni_error)
-                    return ssl_error(env, state->sni_error);
+                    err_term = ssl_error(env, state->sni_error);
                 else
-                    return ssl_error(env, "SSL_do_handshake failed");
+                    err_term = ssl_error(env, "SSL_do_handshake failed");
+                return return_read_write(env, state, bytes_to_read, err_term);
             }
         }
         if (!SSL_is_init_finished(state->ssl)) {
@@ -1119,7 +1130,7 @@ loop_nif(ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[]) {
             if (res == 2) {
                 return err_term;
             }
-            return return_read_write(env, state, bytes_to_read);
+            return return_read_write(env, state, bytes_to_read, enif_make_atom(env, "init"));
         }
     }
 
@@ -1130,7 +1141,7 @@ loop_nif(ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[]) {
 
     if (res <= 0)
         res = SSL_get_error(state->ssl, res);
-    return return_read_write(env, state, bytes_to_read);
+    return return_read_write(env, state, bytes_to_read, enif_make_atom(env, "ok"));
 }
 
 static ERL_NIF_TERM get_verify_result_nif(ErlNifEnv *env, int argc,
@@ -1163,7 +1174,6 @@ static ERL_NIF_TERM get_peer_certificate_nif(ErlNifEnv *env, int argc,
     X509 *cert = NULL;
     state_t *state = NULL;
     int rlen;
-    ErlNifBinary output;
 
     if (argc != 1)
         return enif_make_badarg(env);
@@ -1188,14 +1198,16 @@ static ERL_NIF_TERM get_peer_certificate_nif(ErlNifEnv *env, int argc,
     }
     rlen = i2d_X509(cert, NULL);
     if (rlen >= 0) {
-        if (!enif_alloc_binary(rlen, &output)) {
+        ERL_NIF_TERM bin;
+        unsigned char *buf = enif_make_new_binary(env, rlen, &bin);
+        if (!buf) {
             enif_mutex_unlock(state->mtx);
             return ERR_T(enif_make_atom(env, "enomem"));
         }
-        i2d_X509(cert, &output.data);
+        i2d_X509(cert, &buf);
         X509_free(cert);
         enif_mutex_unlock(state->mtx);
-        return OK_T(enif_make_binary(env, &output));
+        return OK_T(bin);
     } else {
         X509_free(cert);
         enif_mutex_unlock(state->mtx);
@@ -1362,6 +1374,10 @@ static ERL_NIF_TERM tls_get_peer_finished_nif(ErlNifEnv *env, int argc, const ER
     if (!enif_get_resource(env, argv[0], tls_state_t, (void *) &state))
         return enif_make_badarg(env);
 
+#ifdef TLS1_3_VERSION
+    if (SSL_version(state->ssl) >= TLS1_3_VERSION)
+        return ERR_T(enif_make_atom(env, "undefined"));
+#endif
 	/* OpenSSL does not offer an API to directly get the length of the
 	 * expected TLS Finished message, so just do a dummy call to grab this
 	 * information to allow caller to do an allocation with a correct size.
@@ -1383,6 +1399,11 @@ static ERL_NIF_TERM tls_get_finished_nif(ErlNifEnv *env, int argc, const ERL_NIF
     if (!enif_get_resource(env, argv[0], tls_state_t, (void *) &state))
         return enif_make_badarg(env);
 
+#ifdef TLS1_3_VERSION
+    if (SSL_version(state->ssl) >= TLS1_3_VERSION)
+        return ERR_T(enif_make_atom(env, "undefined"));
+#endif
+
     ERL_NIF_TERM bin;
     size_t len = SSL_get_finished(state->ssl, NULL, 0);
     if (len == 0)
@@ -1395,6 +1416,30 @@ static ERL_NIF_TERM tls_get_finished_nif(ErlNifEnv *env, int argc, const ERL_NIF
     return OK_T(bin);
 }
 
+static ERL_NIF_TERM get_tls_cb_exporter_nif(ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[]) {
+    state_t *state = NULL;
+    if (!enif_get_resource(env, argv[0], tls_state_t, (void *) &state))
+        return enif_make_badarg(env);
+
+#ifdef TLS1_3_VERSION
+    ERL_NIF_TERM bin;
+    unsigned char *buf = enif_make_new_binary(env, 32, &bin);
+    if (!buf)
+        return ERR_T(enif_make_atom(env, "enomem"));
+
+    if (SSL_version(state->ssl) < TLS1_3_VERSION && SSL_get_extms_support(state->ssl) <= 0)
+        return ERR_T(enif_make_atom(env, "undefined"));
+
+    if (SSL_export_keying_material(state->ssl, buf, 32,
+         "EXPORTER-Channel-Binding", 24, NULL, 0, 1) != 1)
+        return ERR_T(enif_make_atom(env, "undefined"));
+
+    return OK_T(bin);
+#else
+    return ERR_T(enif_make_atom(env, "undefined"));
+#endif
+}
+
 static ERL_NIF_TERM set_fips_mode_nif(ErlNifEnv *env, int argc,
                                       const ERL_NIF_TERM argv[]) {
   int ret = 1;
@@ -1415,8 +1460,15 @@ static ERL_NIF_TERM set_fips_mode_nif(ErlNifEnv *env, int argc,
 
   if (ret != 1)
     return ssl_error(env, "FIPS_mode_set() failed");
-#elif __GNUC__
-#warning OpenSSL 3 FIPS support not implemented
+#else
+  int fips_mode = EVP_default_properties_is_fips_enabled(NULL);
+
+  if ((fips_mode == 0 && enable != 0) ||
+      (fips_mode != 0 && enable == 0))
+    ret = EVP_default_properties_enable_fips(NULL, enable);
+
+  if (ret != 1)
+    return ssl_error(env, "FIPS_mode_set() failed");
 #endif
 
   return enif_make_atom(env, "ok");
@@ -1428,15 +1480,149 @@ static ERL_NIF_TERM get_fips_mode_nif(ErlNifEnv *env, int argc,
 #if OPENSSL_VERSION_NUMBER < 0x30000000L
   const char *ret = FIPS_mode() ? "true" : "false";
 #else
-#if __GNUC__
-#warning OpenSSL 3 FIPS support not implemented
-#endif
-  static const char *ret = "false";
+  const char *ret = EVP_default_properties_is_fips_enabled(NULL) ? "true" : "false";
 #endif
 
   return enif_make_atom(env, ret);
 }
 
+static ERL_NIF_TERM p12_to_pem_nif(ErlNifEnv *env, int argc,
+                                   const ERL_NIF_TERM argv[]) {
+    ErlNifBinary p12_bin, pass_bin;
+    cert_info_t *info = NULL;
+    cert_info_t *old_info = NULL;
+
+    if (argc != 2)
+        return enif_make_badarg(env);
+
+    if (!enif_inspect_iolist_as_binary(env, argv[0], &p12_bin))
+        return enif_make_badarg(env);
+    if (!enif_inspect_iolist_as_binary(env, argv[1], &pass_bin))
+        return enif_make_badarg(env);
+
+    PKCS12 *p12 = NULL;
+    EVP_PKEY *pkey = NULL;
+    X509 *cert;
+    BIO *bio;
+    ERL_NIF_TERM res;
+
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+    OSSL_LIB_CTX *lib_ctx = OSSL_LIB_CTX_new();
+    OSSL_PROVIDER *prov_def, *prov_leg;
+    if (lib_ctx) {
+        prov_def = OSSL_PROVIDER_load(lib_ctx, "default");
+        prov_leg = OSSL_PROVIDER_load(lib_ctx, "legacy");
+    } else {
+        return ERR_T(enif_make_atom(env, "enomem"));
+    }
+
+    p12 = PKCS12_init_ex(NID_pkcs7_data, lib_ctx, NULL);
+    if (!p12) {
+        res = ERR_T(enif_make_atom(env, "enomem"));
+        goto clean;
+    }
+#endif
+
+    pkey = EVP_PKEY_new();
+    if (!pkey) {
+        res = ERR_T(enif_make_atom(env, "enomem"));
+        goto clean;
+    }
+
+    cert = X509_new();
+    if (!cert) {
+        res = ERR_T(enif_make_atom(env, "enomem"));
+        goto clean;
+    }
+    bio = BIO_new(BIO_s_mem());
+    if (!bio) {
+        res = ERR_T(enif_make_atom(env, "enomem"));
+        goto clean;
+    }
+
+
+    BIO* input = BIO_new_mem_buf(p12_bin.data, p12_bin.size);
+    if (!input) {
+        res = ERR_T(enif_make_atom(env, "enomem"));
+        goto clean;
+    }
+
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+    PKCS12 *p12_tmp = p12;
+    if (!d2i_PKCS12_bio(input, &p12_tmp)) {
+        BIO_free(input);
+        res = ssl_error(env, "p12_to_pem:decode");
+        goto clean;
+    }
+    BIO_free(input);
+#else
+    p12 = d2i_PKCS12_bio(input, NULL);
+    BIO_free(input);
+    if (!p12) {
+        res = ssl_error(env, "p12_to_pem:decode");
+        goto clean;
+    }
+#endif
+    if (!PKCS12_parse(p12, pass_bin.data, &pkey, &cert, NULL)) {
+        unsigned long err = ERR_peek_error();
+        if (ERR_GET_LIB(err) == ERR_LIB_PKCS12 && ERR_GET_REASON(err) == PKCS12_R_MAC_VERIFY_FAILURE) {
+            res =  ERR_T(enif_make_atom(env, "bad_pass"));
+            goto clean;
+        }
+        res = ssl_error(env, "p12_to_pem:parse");
+        goto clean;
+    }
+
+    if (!PEM_write_bio_PrivateKey(bio, pkey, NULL, NULL, 0, 0, NULL)) {
+        res =  ssl_error(env, "p12_to_pem:privkey");
+        goto clean;
+    }
+
+    ERL_NIF_TERM pkey_bin;
+    size_t bio_size = BIO_ctrl_pending(bio);
+    unsigned char *bin_buf = enif_make_new_binary(env, bio_size, &pkey_bin);
+    if (!bin_buf) {
+        res = ERR_T(enif_make_atom(env, "enomem"));
+        goto clean;
+    }
+    BIO_read(bio, bin_buf, bio_size);
+    BIO_reset(bio);
+
+    if (!PEM_write_bio_X509(bio, cert)) {
+        res =  ssl_error(env, "p12_to_pem:cert");
+        goto clean;
+    }
+
+    ERL_NIF_TERM cert_bin;
+    bio_size = BIO_ctrl_pending(bio);
+    bin_buf = enif_make_new_binary(env, bio_size, &cert_bin);
+    if (!bin_buf) {
+        res = ERR_T(enif_make_atom(env, "enomem"));
+        goto clean;
+    }
+    BIO_read(bio, bin_buf, bio_size);
+    res = OK_T(enif_make_tuple2(env, pkey_bin, cert_bin));
+
+clean:
+    PKCS12_free(p12);
+    EVP_PKEY_free(pkey);
+    X509_free(cert);
+    BIO_free(bio);
+
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+    if (lib_ctx) {
+        OSSL_PROVIDER_unload(prov_leg);
+        OSSL_PROVIDER_unload(prov_def);
+        OSSL_LIB_CTX_free(lib_ctx);
+    }
+#endif
+
+    ERR_clear_error();
+
+    return res;
+}
+
+
 static ErlNifFunc nif_funcs[] =
         {
                 {"open_nif",                  10, open_nif},
@@ -1451,8 +1637,10 @@ static ErlNifFunc nif_funcs[] =
                 {"get_negotiated_cipher_nif", 1,  get_negotiated_cipher_nif},
                 {"tls_get_peer_finished_nif", 1,  tls_get_peer_finished_nif},
                 {"tls_get_finished_nif",      1,  tls_get_finished_nif},
+                {"get_tls_cb_exporter_nif",   1,  get_tls_cb_exporter_nif},
                 {"set_fips_mode_nif",         1,  set_fips_mode_nif},
-                {"get_fips_mode_nif",         0,  get_fips_mode_nif}
+                {"get_fips_mode_nif",         0,  get_fips_mode_nif},
+                {"p12_to_pem_nif",            2,  p12_to_pem_nif}
         };
 
 ERL_NIF_INIT(fast_tls, nif_funcs, load, NULL, NULL, unload)
