@@ -5,7 +5,7 @@
 %%% Created : 24 Jul 2004 by Alexey Shchepin <alexey@process-one.net>
 %%%
 %%%
-%%% Copyright (C) 2002-2022 ProcessOne, SARL. All Rights Reserved.
+%%% Copyright (C) 2002-2024 ProcessOne, SARL. All Rights Reserved.
 %%%
 %%% Licensed under the Apache License, Version 2.0 (the "License");
 %%% you may not use this file except in compliance with the License.
@@ -29,7 +29,8 @@
 -export([open_nif/10, loop_nif/4, get_peer_certificate_nif/1,
          get_verify_result_nif/1, invalidate_nif/1,
          get_negotiated_cipher_nif/1, set_fips_mode_nif/1,
-         get_fips_mode_nif/0]).
+         get_fips_mode_nif/0, get_tls_cb_exporter_nif/1,
+         p12_to_pem_nif/2]).
 
 -export([tcp_to_tls/2,
          tls_to_tcp/1, send/2, recv/2, recv/3, recv_data/2,
@@ -39,7 +40,8 @@
          get_verify_result/1, get_cert_verify_string/2,
          add_certfile/2, get_certfile/1, delete_certfile/1,
          clear_cache/0, get_negotiated_cipher/1,
-         get_tls_last_message/2, set_fips_mode/1, get_fips_mode/0]).
+         get_tls_last_message/2, set_fips_mode/1, get_fips_mode/0,
+         get_tls_cb_exporter/1, p12_to_pem/2, finish_handshake/2]).
 
 -ifdef(TEST).
 -include_lib("eunit/include/eunit.hrl").
@@ -113,12 +115,18 @@ tls_get_peer_finished_nif(_Port) ->
 tls_get_finished_nif(_Port) ->
     erlang:nif_error({nif_not_loaded, ?MODULE}).
 
+get_tls_cb_exporter_nif(_Port) ->
+    erlang:nif_error({nif_not_loaded, ?MODULE}).
+
 set_fips_mode_nif(_Flag) ->
     erlang:nif_error({nif_not_loaded, ?MODULE}).
 
 get_fips_mode_nif() ->
     erlang:nif_error({nif_not_loaded, ?MODULE}).
 
+p12_to_pem_nif(_P12Data, _Pass) ->
+    erlang:nif_error({nif_not_loaded, ?MODULE}).
+
 %%% --------------------------------------------------------
 %%% The call-back functions.
 %%% --------------------------------------------------------
@@ -228,16 +236,19 @@ loop(#tlssock{tcpsock = TCPSocket,
     try loop_nif(Port, ToSend, Received, Length) of
         {error, _} = Err ->
             Err;
-        {<<>>, Decrypted} ->
+        {Tag, <<>>, Decrypted} when Tag == ok; Tag == init ->
             {ok, <<DecBuf/binary, Decrypted/binary>>};
-        {ToWrite, Decrypted} ->
+        {Tag, ToWrite, Decrypted} when Tag == ok; Tag == init ->
             case gen_tcp:send(TCPSocket, ToWrite) of
                 ok ->
                     loop(Socket, <<>>, <<>>, <<DecBuf/binary, Decrypted/binary>>,
                          Length - byte_size(Decrypted));
                 {error, _} = Err ->
                     Err
-            end
+            end;
+        {{error, _} = Err, ToWrite, _} ->
+            _ = gen_tcp:send(TCPSocket, ToWrite),
+            Err
     catch error:badarg ->
         {error, einval}
     end.
@@ -271,6 +282,38 @@ recv_and_loop(#tlssock{tcpsock = TCPSocket} = Socket,
             end
     end.
 
+-spec finish_handshake(tls_socket(), timeout()) ->
+    {error, inet:posix() | binary()} | ok.
+finish_handshake(#tlssock{tcpsock = TCPSocket, tlsport = Port}, Timeout) ->
+    OurLoop = fun OurLoop(Received) ->
+        try loop_nif(Port, <<>>, Received, 0) of
+            {error, _} = Err ->
+                Err;
+            {Tag, ToWrite, <<>>} when Tag == ok; Tag == init ->
+                case gen_tcp:send(TCPSocket, ToWrite) of
+                    ok when Tag == init ->
+                        case gen_tcp:recv(TCPSocket, 0, Timeout) of
+                            {ok, Received2} ->
+                                OurLoop(Received2);
+                            {error, _} = Err ->
+                                Err
+                        end;
+                    ok ->
+                        ok;
+                    {error, _} = Err ->
+                        Err
+                end;
+            {Tag, _ToWrite, _Data} when Tag == ok; Tag == init ->
+                {error, too_much_data_received};
+            {{error, _} = Err, ToWrite, _} ->
+                _ = gen_tcp:send(TCPSocket, ToWrite),
+                Err
+        catch error:badarg ->
+            {error, einval}
+        end
+    end,
+    OurLoop(<<>>).
+
 -spec send(tls_socket(), binary()) ->
     ok | {error, inet:posix() | binary() | timeout}.
 send(Socket, Packet) ->
@@ -340,6 +383,10 @@ get_tls_last_message(peer, #tlssock{tlsport = Port}) ->
 get_tls_last_message(self, #tlssock{tlsport = Port}) ->
     tls_get_finished_nif(Port).
 
+-spec get_tls_cb_exporter(tls_socket()) -> {ok, binary()} | {error, term()}.
+get_tls_cb_exporter(#tlssock{tlsport = Port}) ->
+    get_tls_cb_exporter_nif(Port).
+
 -spec get_verify_result(tls_socket()) -> byte().
 get_verify_result(#tlssock{tlsport = Port}) ->
     {ok, Res} = get_verify_result_nif(Port),
@@ -393,6 +440,11 @@ set_fips_mode(false) ->
 get_fips_mode() ->
     get_fips_mode_nif().
 
+%% @doc Converts PKCS12 certificate to PEM encoded private key and certificate
+-spec p12_to_pem(binary(), binary()) -> {ok, {binary(), binary()}} | {error, atom() | binary()}.
+p12_to_pem(P12Data, Pass) ->
+    p12_to_pem_nif(P12Data, <<Pass/binary, 0>>).
+
 cert_verify_code(0)  -> <<"ok">>;
 cert_verify_code(2) ->
     <<"unable to get issuer certificate">>;
